/*

What is screenbox?

	A screenbox is a HUD object that snaps into a client's view.
	It is anchored to a world position,
		so it has to reposition itself whenever the camera moves.
	If it sticks out of the camera,
		it snaps back into view so it stays completely visible.

Example use:

	A sign on the map can be read by players.
		Upon reading the sign, a textbox appears above it.
		When the player moves around, the textbox stays above the sign.
		If the player is reading the sign from the south,
			the top of the textbox might be cut off,
			but it isn't.

Customizations:

	SetSize(width, height)
		Size of the box.
		(in pixels)

	SetCenter(X, Y)
		World coordinate of the box's center.
		(in pixels)

	SetMargin(left, bottom, right, top)
		Extra spacing between the edges of the box and its container.
		(in pixels)

	SetLimit(num)
		Distance past the screen edge before the box disappears.
		(in pixels)

	SetShadow(color, x-offset, y-offset)
		Color of the shadow;
		Offset of the shadow from the box's content.

*/

#include "..\..\lib\vector2.dm"
#include "..\..\lib\aabb.dm"

#ifndef TILE_WIDTH
#define TILE_WIDTH 32
#endif

#ifndef TILE_HEIGHT
#define TILE_HEIGHT 32
#endif

#if DM_VERSION < 509 && defined(DEBUG)
client
	var
		bounds[] = list(-1#INF, -1#INF, 0, 0)
		bound_x = -1#INF
		bound_y = -1#INF
		bound_width = 0 
		bound_height = 0
#endif

screenbox
	parent_type = /obj
	screen_loc = "1,1"
	mouse_opacity = FALSE
	alpha = 0

	var
		client/client
		vector2/center

		aabb/aabb
		half_width
		half_height

		dirty

		shadow_color
		vector2/shadow_offset
		obj/shadow

		aabb/margin
		limit = 1#INF

		update_period
		
		last_bounds[]
		
		is_visible

	New(Client, CenterX, CenterY, Width, Height)
		client = Client
		aabb = new

		SetCenter(CenterX, CenterY)
		SetSize(Width, Height)
		Loop()

	proc
		Loop()
			set waitfor = FALSE
			for()
				update_period = update_period || world.tick_lag
				sleep update_period
				
				if(!client) break
				
				dirty = dirty || !last_bounds
				
				if(!dirty)
					// check if the viewport moved
					var client_bounds[] = client.bounds
					for(var/n in 1 to client_bounds.len)
						if(last_bounds[n] != client_bounds[n])
							dirty = TRUE
							break
				
				dirty && Reposition()
				

		SetSize(Width, Height)
			half_width  = Width/2
			half_height = Height/2
			SetCenter(center.x, center.y)

		SetCenter(X, Y)
			center = vector2(X, Y)
			aabb.Set(
				X - half_width - TILE_WIDTH/2, Y - half_height - TILE_HEIGHT/2,
				X + half_width - TILE_WIDTH/2, Y + half_height - TILE_HEIGHT/2)
			dirty = TRUE

		Reposition()
			set waitfor = FALSE
			
			dirty = FALSE
			last_bounds = client.bounds
			
			var
				view_min_x = client.bound_x
				view_min_y = client.bound_y
				view_max_x = view_min_x + client.bound_width
				view_max_y = view_min_y + client.bound_height
				snapped_x = aabb.min_x
				snapped_y = aabb.min_y
			
			// enforce the margin, if any
			if(margin)
				view_min_x += margin.min_x
				view_max_x -= margin.max_x
				view_min_y += margin.min_y
				view_max_y -= margin.max_y

			// snap the box into view
			if		(aabb.min_x < view_min_x) snapped_x += view_min_x - aabb.min_x
			else if	(aabb.max_x > view_max_x) snapped_x += view_max_x - aabb.max_x
			if		(aabb.min_y < view_min_y) snapped_y += view_min_y - aabb.min_y
			else if	(aabb.max_y > view_max_y) snapped_y += view_max_y - aabb.max_y
			
			if(abs(snapped_x - aabb.min_x) > limit || abs(snapped_y - aabb.min_y) > limit)
				is_visible && Hide()
				
			else
				!is_visible && Show()
				var matrix/m = transform
				m.c = snapped_x - client.bound_x
				m.f = snapped_y - client.bound_y
				transform = m

		Hide()
			OnHide()
			is_visible = FALSE
			client.screen -= src

		Show()
			is_visible = TRUE
			client.screen += src
			OnShow()

		OnHide()
			animate(src, time = 2, alpha = 0)
			sleep 2
		
		OnShow()
			animate(src, time = 2, alpha = 255)

		SetShadow(Color, OffsetX, OffsetY)
			if(!shadow)
				shadow = new
				shadow.icon_state = "rect"
				shadow.layer = FLOAT_LAYER
			else underlays -= shadow
			shadow.color = Color
			shadow.transform = matrix(
				aabb.width/32, 	0, 				aabb.width/2  + OffsetX - 16, 
				0, 				aabb.height/32, aabb.height/2 + OffsetY - 16)
			underlays += shadow

		SetMargin(Left, Bottom, Right, Top)
			if(!margin) margin = new (Left, Bottom, Right, Top)
			else margin.Set(Left, Bottom, Right, Top)
			dirty = TRUE

		SetLimit(Limit)
			limit = Limit
			dirty = TRUE
